from pyinfra.api import operation
from pyinfra import host
from pyinfra_net.facts import switch
from pyinfra_net.drivers import get_driver

from typing import Optional, Generator, Dict, List, Tuple
from ipaddress import ip_network, ip_address


def _sanitize_vlan_name(name: str) -> str:
    return name[0:32]


@operation()
def vlan(vlan_id: int, name: Optional[str] = "", present: Optional[bool] = True, debug: Optional[bool] = False) -> Generator[str, None, None]: 
    existing_vlans = host.get_fact(switch.Vlans)
    name = _sanitize_vlan_name(name)
    
    drv = get_driver(host.data.get("device_type"), debug)

    if present and vlan_id not in existing_vlans:
        yield from drv.create_vlan(vlan_id, name)
        yield from drv.save()
        return
    elif present and vlan_id in existing_vlans and name != existing_vlans[vlan_id]:
        yield from drv.change_vlan_name(vlan_id, name)
        yield from drv.save()
        return
    elif not present and vlan_id in existing_vlans:
        yield from drv.delete_vlan(vlan_id)
        yield from drv.save()
        return


@operation()
def vlans(vlans: Dict[int, str], debug: Optional[bool] = False):
    vlans = { int(vlan_id): _sanitize_vlan_name(name) for vlan_id, name in vlans.items() }
    existing_vlans = host.get_fact(switch.Vlans)
    
    drv = get_driver(host.data.get("device_type"), debug)

    to_delete = [vlan_id for vlan_id in existing_vlans if vlan_id not in vlans]
    to_create = {vlan_id: vlans[vlan_id] for vlan_id in vlans if vlan_id not in existing_vlans}
    to_rename = {vlan_id: vlans[vlan_id] for vlan_id in vlans if vlan_id in existing_vlans and vlans[vlan_id] != existing_vlans[vlan_id]}

    for vlan_id in to_delete:
        yield from drv.delete_vlan(vlan_id)

    for vlan_id, name in to_create.items():
        yield from drv.create_vlan(vlan_id, name)

    for vlan_id, name in to_rename.items():
        yield from drv.rename_vlan(vlan_id, name)

    if to_delete or to_create or to_rename:
        yield from drv.save()


@operation()
def route(destination: str, gateway: str, present: Optional[bool] = True, debug: Optional[bool] = False):
    existing_routes = host.get_fact(switch.Routes)
    destination = ip_network(destination)
    gateway = ip_address(gateway)
    route = (destination, gateway)

    drv = get_driver(host.data.get("device_type"), debug)

    if present and route not in existing_routes:
        yield from drv.create_route(destination, gateway)
        yield from drv.save()
        return
    elif not present and route in existing_routes:
        yield from drv.delete_route(destination, gateway)
        yield from drv.save()
        return


@operation()
def routes(routes: List[Tuple[str, str]], debug: Optional[bool] = False):
    existing_routes = host.get_fact(switch.Routes)
    desired_routes = [ (ip_network(destination), ip_address(gateway)) for destination, gateway in routes ]

    drv = get_driver(host.data.get("device_type"), debug)

    to_delete = [ route for route in existing_routes if route not in desired_routes ]
    to_create = [ route for route in desired_routes if route not in existing_routes ]

    for destination, gateway in to_delete:
        yield from drv.delete_route(destination, gateway)

    for destination, gateway in to_create:
        yield from drv.create_route(destination, gateway)

    if to_delete or to_create:
        yield from drv.save()


from __future__ import annotations

from .base import NetworkDeviceDriver, register_driver
from typing import Generator, Iterable, Dict, Tuple
from ipaddress import ip_network, ip_address


@register_driver("netgear_prosafe")
class NetgearProsafeDriver(NetworkDeviceDriver):
    @property
    def comment_symbol(self) -> str:
        return "!"
    
    @property
    def expect_string(self) -> str:
        return r"\(.*\)\s*(\((Vlan|Config)\))?#"
    
    def show_vlans(self) -> str:
        return "show vlan"
    
    def show_routes(self) -> str:
        return "show ip route static"
    
    def process_vlans(self, output: Iterable[str]) -> Dict[int, str]:
        vlans = {}
        for line in output:
            
            if line == "":
                continue
            
            if not line[0].isdigit():
                continue
            
            parts = line.split()
            vlan_id = int(parts[0])
            vlan_name = parts[1]

            if vlan_name == "VLAN"+str(vlan_id):
                vlan_name = ""
            
            if vlan_id == 1:
                continue
            
            vlans[vlan_id] = vlan_name

        return vlans
    
    def process_routes(self, output: Iterable[str]) -> Iterable[Tuple[ip_network, ip_address]]:
        for line in output:

            if line == "":
                continue
            
            if not line[0] == "S":
                continue
            
            parts = line.split()
            
            if parts[1] == "0.0.0.0/0":
                continue
            
            destination = ip_network(parts[1])
            gateway = ip_address(parts[4].replace(",", ""))
            yield (destination, gateway)

    def _create_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        yield "vlan database"
        yield f"vlan {vlan_id}"
        if name:
            yield f"vlan name {vlan_id} {name}"
        yield "exit"

    def _delete_vlan(self, vlan_id: int) -> Generator[str, None, None]:
        yield "vlan database"
        yield f"no vlan {vlan_id}"
        yield "exit"

    def _rename_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        yield "vlan database"
        yield f"vlan name {vlan_id} {name}"
        yield "exit"

    def _create_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        yield "configure"
        yield f"ip route {str(destination.network_address)} {str(destination.netmask)} {str(gateway)}"
        yield "exit"

    def _delete_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        yield "configure"
        yield f"no ip route {str(destination.network_address)} {str(destination.netmask)} {str(gateway)}"
        yield "exit"

    def _save(self) -> Generator[str, None, None]:
        yield "write memory confirm"


from __future__ import annotations

from .base import NetworkDeviceDriver, register_driver

import re
from typing import Generator, Iterable, Dict, Tuple
from ipaddress import ip_network, ip_address

@register_driver("mellanox")
class MellanoxDriver(NetworkDeviceDriver):
    @property
    def comment_symbol(self) -> str:
        return "#"
    
    def show_vlans(self) -> str:
        return "show vlan"
    
    def show_routes(self) -> str:
        return "show ip route static"
    
    def process_vlans(self, output: Iterable[str]) -> Dict[int, str]:
        vlans = {}
        
        for line in output:
            
            if line == "":
                continue
            
            if not line[0].isdigit():
                continue
            
            vlan_id = int(line[0:9].strip())
            vlan_name = line[10:36].strip()
            if vlan_id == 1:
                continue
            vlans[vlan_id] = vlan_name

        return vlans

    def process_routes(self, output: Iterable[str]) -> Iterable[Tuple[ip_network, ip_address]]:
        for line in output:

            if not re.match(r'^\s{2}(default|\d+\.\d+\.\d+\.\d+|[0-9a-f]{4}:)', line):
                continue

            parts = re.split(r'\s+', line.strip())
            
            if parts[0] == "default":
                continue
            
            destination_net = parts[0]
            destination_mask = parts[1]
            gateway = parts[2]
            
            destination = ip_network(f"{destination_net}/{destination_mask}")
            gateway = ip_address(gateway)
            yield (destination, gateway)

    def _create_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        yield "configure terminal"
        yield f"vlan {vlan_id}"
        if name:
            yield f"name {name}"
        yield "exit"
        yield "exit"

    def _delete_vlan(self, vlan_id: int) -> Generator[str, None, None]:
        yield "configure terminal"
        yield f"no vlan {vlan_id}"
        yield "exit"

    def _rename_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        yield "configure terminal"
        yield f"vlan {vlan_id}"
        yield f"name {name}"
        yield "exit"
        yield "exit"

    def _create_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        yield "configure terminal"
        yield f"ip route {str(destination)} {str(gateway)}"
        yield "exit"

    def _delete_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        yield "configure terminal"
        yield f"no ip route {str(destination)} {str(gateway)}"
        yield "exit"

    def _save(self) -> Generator[str, None, None]:
        yield "write memory"



from __future__ import annotations

from typing import Optional

from .base import NetworkDeviceDriver, register_driver, get_driver_class

from . import netgear_prosafe
from . import mellanox

def get_driver(name: str, debug: Optional[bool] = False) -> NetworkDeviceDriver:
    try:
        driver = get_driver_class(name)
        instance = driver()
        instance.debug = debug
        return instance
    except KeyError:
        raise NotImplementedError(f"Unsupported device type: {name}")

__all__ = [ 
    "NetworkDeviceDriver",
    "register_driver",
    "get_driver",
    "get_driver_class",
]


from pyinfra.api import FactBase
from pyinfra import host
from pyinfra_net.drivers import get_driver

class Vlans(FactBase):
    def command(self):
        return get_driver(host.data.get("device_type")).show_vlans()

    def process(self, output):
        return get_driver(host.data.get("device_type")).process_vlans(output)

class Routes(FactBase):
    def command(self):
        return get_driver(host.data.get("device_type")).show_routes()

    def process(self, output):
        return list(get_driver(host.data.get("device_type")).process_routes(output))
    


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generator, Iterable, Type, Dict, Optional, Tuple
from ipaddress import ip_network, ip_address


class NetworkDeviceDriver(ABC):
    debug: bool = False

    @abstractmethod
    def show_vlans(self) -> str:
        pass

    @abstractmethod
    def show_routes(self) -> str:
        pass

    @abstractmethod
    def process_vlans(self, output: Iterable[str]) -> Dict[int, str]:
        pass

    @abstractmethod
    def process_routes(self, output: Iterable[str]) -> Iterable[Tuple[ip_network, ip_address]]:
        pass

    @property
    def expect_string(self) -> Optional[str]:
        return None

    def create_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        if vlan_id == 1:
            return
        if self.debug:
            yield from self.comment(self._create_vlan(vlan_id, name))
        else:
            yield from self._create_vlan(vlan_id, name)

    def delete_vlan(self, vlan_id: int) -> Generator[str, None, None]:
        if vlan_id == 1:
            return
        if self.debug:
            yield from self.comment(self._delete_vlan(vlan_id))
        else:
            yield from self._delete_vlan(vlan_id)

    def rename_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        if vlan_id == 1:
            return
        if self.debug:
            yield from self.comment(self._change_vlan_name(vlan_id, name))
        else:
            yield from self._change_vlan_name(vlan_id, name)

    def create_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        if self.debug:
            yield from self.comment(self._create_route(destination, gateway))
        else:
            yield from self._create_route(destination, gateway)

    def delete_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        if self.debug:
            yield from self.comment(self._delete_route(destination, gateway))
        else:
            yield from self._delete_route(destination, gateway)

    def save(self) -> Generator[str, None, None]:
        if self.debug:
            yield from self.comment(self._save())
        else:
            yield from self._save()

    def comment(self, commands: Iterable[str]) -> Generator[str, None, None]:
        for command in commands:
            yield f"{self.comment_symbol} {command}"

    @property
    @abstractmethod
    def comment_symbol(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _create_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def _delete_vlan(self, vlan_id: int) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def _rename_vlan(self, vlan_id: int, name: str) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def _create_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def _delete_route(self, destination: ip_network, gateway: ip_address) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def _save(self) -> Generator[str, None, None]:
        pass


_DRIVER_REGISTRY: Dict[str, Type[NetworkDeviceDriver]] = {}


def register_driver(name: str):
    def decorator(cls: Type[NetworkDeviceDriver]) -> Type[NetworkDeviceDriver]:
        _DRIVER_REGISTRY[name] = cls
        return cls
    
    return decorator


def get_driver_class(name: str) -> Type[NetworkDeviceDriver]:
    try:
        return _DRIVER_REGISTRY[name]
    except KeyError:
        raise NotImplementedError(f"Unsupported device type: {name}")


from netmiko import ConnectHandler
from pyinfra.connectors.base import BaseConnector
from pyinfra.connectors.util import CommandOutput, OutputLine
from pyinfra_net.drivers import get_driver
import click

class NetmikoConnector(BaseConnector):
    handles_execution = True

    @staticmethod
    def make_names_data(hostname):
        yield "@netmiko/{}".format(hostname), {}, []

    def connect(self):
        kwargs = {
            'device_type': self.host.data.get('device_type'),
            'host': self.host.data.get('netmiko_hostname') or self.host.name,
            'username': self.host.data.get('username') or 'admin',
            'password': self.host.data.get('password') or None,
            'port': self.host.data.get('port') or 22,
            'secret': self.host.data.get('enable_password') or None,
            'verbose': False,
        }

        self.connection = ConnectHandler(**kwargs)

    def disconnect(self):
        pass

    def run_shell_command(self, command, print_output=False, print_input=False, **arguments):
        command_str = str(command)
        lines = [ line for line in command_str.splitlines() if line.strip() ]

        if print_input:
            for line in lines:
                click.echo("{0}>>> {1}".format(self.host.print_prefix, line), err=True)

        expect_string = get_driver(self.host.data.get('device_type')).expect_string

        if len(lines) > 1:
            output = self.connection.send_config_set(lines, expect_string=expect_string)
        else:
            output = self.connection.send_command(command_str, expect_string=expect_string)
        
        if print_output and output:
            click.echo("{0}{1}".format(self.host.print_prefix, output), err=True)

        return True, CommandOutput([OutputLine("stdout", line) for line in output.splitlines()])
    
    def put_file(self, filename_or_io, remote_filename, remote_temp_filename, print_output=False, print_input=False, **arguments):
        return False
    
    def get_file(self, filename_or_io, remote_filename, remote_temp_filename, print_output=False, print_input=False, **arguments):
        return False
    
    def check_can_rsync(self):
        raise NotImplementedError("Netmiko connector does not support rsync")